
function _block_read_compact_field(f::Symbol, T, name::Symbol, n::Integer;
                                   protocol::Symbol=:p, index::Symbol=:k, type::Symbol=:t,
                                  )
    lhs = :(($index, $type, $name))
    rhs = :(Thrift2.$f($protocol, $index, $type, $n, Thrift2.thriftfieldtype(Thrift2.ThriftStruct{$T}, Val($n))))
    :($lhs = $rhs)
end

function _block_readshort_compact_allfields_expr(f, T, names; kw...)
    map((n, name) -> _block_read_compact_field(f, T, name, n; kw...), 1:length(names), names)
end

_block_readshort_compact_allfields(T, names; kw...) = quote
    $(_block_readshort_compact_allfields_expr(:readshortfield, T, names; kw...)...)
end |> unblock

_block_read_compact_output(T, names; index::Symbol=:k) = quote
    if $index == 0
        $T($(names...))
    else
        throw(Thrift2.ThriftInvalidFinalReadState($T, $index))
    end
end |> unblock

gencode(::typeof(readshort), ::Type{CompactProtocol}, T, names; index::Symbol=:k, type::Symbol=:t, kw...) = quote
    function Thrift2.readshort(p::Thrift2.CompactProtocol, ::Type{Thrift2.ThriftStruct{$T}},
                               $index::Integer, $type::Integer,
                              )
        $(_block_readshort_compact_allfields(T, names; index, type, kw...))
        $(_block_read_compact_output(T, names; index))
    end
end |> unblock

function _gencode(::typeof(thriftfieldtype), T, n::Integer)
    lhs = :(Thrift2.thriftfieldtype(::Type{Thrift2.ThriftStruct{$T}}, ::Val{$n}))
    rhs = :(Thrift2.thrifttype(fieldtype($T, $n)))
    :($lhs = $rhs)
end

function gencode(::typeof(thriftfieldtype), T, nf::Integer)
    v = map(n -> _gencode(thriftfieldtype, T, n), 1:nf)
    quote $(v...) end |> unblock
end

gencode(::typeof(thrifttype), T) = :(Thrift2.thrifttype(::Type{$T}) = Thrift2.ThriftStruct{$T}())

function _block_readlong_decs(names)
    exs = map(n -> :($n = nothing), names)
    quote $(exs...) end |> unblock
end

function _block_readlong_compact_allfields(T, names; index::Symbol=:k, kw...)
    rs = _block_readshort_compact_allfields_expr(:readlongfield, T, names; kw...)
    # the read loop is finite so that it doesn't hang forever if something goes wrong
    # thrift only supports typemax(Int16) fields, so this is perfectly valid
    quote
        for _ ∈ 1:typemax(Int16)
            $(rs...)
            $index == 0 && break
        end
    end |> unblock
end

gencode(::typeof(readlong), ::Type{CompactProtocol}, T, names; index::Symbol=:k, type::Symbol=:t, kw...) = quote
    function Thrift2.readlong(p::Thrift2.CompactProtocol, ::Type{Thrift2.ThriftStruct{$T}},
                              $index::Integer, $type::Integer,
                             )
        $(_block_readlong_decs(names))
        $(_block_readlong_compact_allfields(T, names; index, type, kw...))
        $T($(names...))
    end
end |> unblock

gencode(::typeof(read), ::Type{CompactProtocol}, T) = quote
    function Base.read(p::Thrift2.CompactProtocol, ::Type{Thrift2.ThriftStruct{$T}})
        (l, k, t) = Thrift2.readinitfieldheader(p)
        l ? Thrift2.readlong(p, Thrift2.ThriftStruct{$T}, k, t) : Thrift2.readshort(p, Thrift2.ThriftStruct{$T}, k, t)
    end
    Base.read(p::Thrift2.CompactProtocol, ::Type{$T}) = read(p, Thrift2.ThriftStruct{$T})
end |> unblock

gencode(::typeof(==), T) = quote
    Base.:(==)(a::$T, b::$T) = all(j -> getfield(a, j) == getfield(b, j), 1:fieldcount($T))
end |> unblock

genallcode(T, names) = quote
    $(gencode(==, T))
    $(gencode(thriftfieldtype, T, length(names)))
    $(gencode(thrifttype, T))
    $(gencode(readshort, CompactProtocol, T, names))
    $(gencode(readlong, CompactProtocol, T, names))
    $(gencode(read, CompactProtocol, T))
end |> unblock

function _extract_field_names(fields)
    names = Vector{Union{Symbol,Expr}}(undef, length(fields))
    for j ∈ 1:length(fields)
        ϕ = fields[j]
        if @capture(ϕ, name_::S_) || @capture(ϕ, name_::S_ = default_)
            names[j] = name
        else
            throw(ArgumentError("invalid thrift struct field definition: $ϕ"))
        end
    end
    names
end

"""
    @thriftstruct struct_declaration

A macro which declares a struct and generates all code necessary for a struct that can be serialized using the
thrift protocol.  The fields of this struct must be of types which are accommodated by the protocol.
Optional fields of type `T` must be declared with type `Union{Nothing,T}` where `T` is the native
Julia type of the field.

This macro will in turn use the `@kwdef` macro which creates a keyword argument constructor and allows
field defaults to be specified in the struct declaration itself.  Optional fields should have the default value
`nothing` to comply with the protocol.

The declared type must be immutable and have no parameters, but it is allowed to be of any abstract type.

## Examples
```julia
@thriftstruct struct TestStruct <: AbstractTestType
    a::Vector{UInt8}  # mandatory bytes field
    b::Union{Nothing,Float64} = nothing  # optional floats field
    c::Nothing = nothing  # field 3 not defined in thrift schema, placeholder needed
    d::Union{Nothing,Vector{String}} = nothing  # optional arrays are just like any other optional
end
```
"""
macro thriftstruct(tdec)
    if !(@capture(tdec, struct T_ <: ST_ fields__ end) || @capture(tdec, struct T_ fields__ end))
        error("@thriftstruct must have struct definition as argument")
    end
    names = _extract_field_names(fields)
    expr = genallcode(T, names)
    # note that @kwdef was unexported before 1.9
    quote
        Base.@kwdef $tdec
        $expr
    end |> unblock |> esc
end
