

"""
    unzigzag(x)

Invert the [zig-zag encoding](https://en.wikipedia.org/wiki/Variable-length_quantity#Zigzag_encoding)
of an integer `x`.  That is, `unzigzag(zigzag(x)) == x`.
"""
unzigzag(x::Integer) = -(x & Int64(1)) ⊻ (x >>> 1)

"""
    readvarint(io, ::Type{T}) where {T<:Integer}

Read a [varint encoded](https://en.wikipedia.org/wiki/Variable-length_quantity#Group_Varint_Encoding)
integer from the `io` stream.  The result is of integer type `T`.
"""
function readvarint(io::IO, ::Type{T}) where {T<:Integer}
    o = zero(T)
    n = 0
    b = UInt8(0x80)
    while (b & UInt8(0x80)) ≠ 0
        b = read(io, UInt8)
        o |= convert(T, b & 0x7f) << 7n
        n += 1
    end
    o
end

Base.read(p::ThriftProtocol, t::ThriftType) = read(p, typeof(t))

function Base.read(p::CompactProtocol, ::Type{T}) where {T<:ThriftInteger}
    jt = juliatype(T)
    convert(jt, unzigzag(readvarint(p.io, jt)))
end
Base.read(p::CompactProtocol, ::Type{<:ThriftFloat64}) = reinterpret(Float64, ntoh(read(p.io, UInt64)))

Base.read(p::CompactProtocol, ::Type{ThriftEnum{T}}) where {T<:Enum} = T(read(p, ThriftInt32))

function _read_setlike_elements(p::CompactProtocol, s::Integer, ::Type{T}) where {T<:ThriftType}
    o = Vector{juliatype(T)}(undef, s)
    for j ∈ 1:s
        o[j] = read(p, T)
    end
    o
end
function _read_setlike_elements(p::CompactProtocol, s::Integer, t::Integer)
    _read_setlike_elements(p, s, decodetype(CompactProtocol, t))
end

function _read_setlike_preamble(p::CompactProtocol)
    h = read(p.io, UInt8)
    s = h >> 4
    t = 0x0f & h
    s == 0x0f && (s = readvarint(p.io, Int32))  # special code for reading size separately
    (s, t)
end

# this method for when we know eltype from schema
function Base.read(p::CompactProtocol, ::Type{<:ThriftSetlike{T}}) where {T}
    (s, _) = _read_setlike_preamble(p)
    _read_setlike_elements(p, s, T)
end

# this method for when eltype is unknown
function Base.read(p::CompactProtocol, ::Type{ThriftSetlike})
    (s, t) = _read_setlike_preamble(p)
    _read_setlike_elements(p, s, t)
end

Base.read(p::CompactProtocol, ::Type{<:ThriftBytes}) = read(p.io, readvarint(p.io, Int32))

Base.read(p::CompactProtocol, ::Type{<:ThriftString}) = String(read(p, ThriftBytes))

function Base.read(p::CompactProtocol, ::Type{<:ThriftMap{K,V}}) where {K,V}
    o = OrderedDict{juliatype(K),juliatype(V)}()
    n = readvarint(p.io, Int32)
    n == 0 && return 0
    sizehint!(o, n)
    kv = read(p.io, UInt8)  # these give types, but we only support getting them from schema
    for j ∈ 1:n
        k = read(p, K)  # these are separate so that the order is really obvious
        v = read(p, V)
        o[k] = v
    end
    o
end

_readfield_split_header(a::Integer) = (Int(a >>> 4), Int(a & 0x0f))

_readlong_parse_header(p::CompactProtocol, a::Integer, b::Integer, j::Integer) = (Int(unzigzag(read(p.io, UInt16))), b)

_readshort_parse_header(p::CompactProtocol, a::Integer, b::Integer, j::Integer) = (a+j, b)

"""
    readinitfieldheader(p::CompactProtocol, j::Integer=0)

Read the initial field header of a struct serialized according to the thrift compact protocol.
This function is used in the generated reading code for a thrift struct.
"""
function readinitfieldheader(p::CompactProtocol, j::Integer=0)::Tuple{Bool,Int,Int}
    a = read(p.io, UInt8)
    a == 0 && return (false, 0, -1)  # struct is empty
    (a, b) = _readfield_split_header(a)
    long = a == 0
    (k, tid) = (long ? _readlong_parse_header : _readshort_parse_header)(p, a, b, j)
    (long, k, tid)
end

"""
    readshortfieldheader(p::CompactProtocol, j::Integer=0)

Read the field header of a thrift struct, serialized in the short form (for structs with
≤ 15 fields).
This function is used in the generated reading code for a thrift struct.
"""
function readshortfieldheader(p::CompactProtocol, j::Integer=0)::Tuple{Int,Int}
    a = read(p.io, UInt8)
    a == 0 && return (0, -1)
    (a, b) = _readfield_split_header(a)
    _readshort_parse_header(p, a, b, j)
end

"""
    readlongfieldheader(p::CompactProtocol, j::Integer=0)

Read the field header of a thrift struct, serialized in the long form (for structs with
> 15 fields).
This function is used in the generated reading code for a thrift struct.
"""
function readlongfieldheader(p::CompactProtocol, j::Integer=0)::Tuple{Int,Int}
    a = read(p.io, UInt8)
    a == 0 && return (0, -1)
    (a, b) = _readfield_split_header(a)
    _readlong_parse_header(p, a, b, j)
end

function _readfield(f, p::CompactProtocol, k::Integer, t::Integer, id::Integer, T)
    o = k == id ? read(p, T) : nothing
    isnothing(o) || ((k, t) = f(p, k))
    (k, t, o)
end
function _readfield(f, p::CompactProtocol, k::Integer, t::Integer, id::Integer, ::ThriftBool)
    o = k == id ? (t == 1) : nothing
    isnothing(o) || ((k, t) = f(p, k))
    (k, t, o)
end

"""
    readshortfield(p::CompactProtocol, k::Integer, t::Integer, id::Integer, T)

Read a field of type `T` of a struct with field id `id` of a thrift struct, serliazed in
the short form (for structs with < 15 fields).
This function is used in the generated reading code for a thrift struct.
"""
function readshortfield(p::CompactProtocol, k::Integer, t::Integer, id::Integer, T) 
    _readfield(readshortfieldheader, p, k, t, id, T)
end

"""
    readlongfield(p::CompactProtocol, k::Integer, t::Integer, id::Integer, T)

Read a field of type `T` of a struct with field id `id` of a thrift struct, serliazed in
the long form (for structs with ≥ 15 fields).
This function is used in the generated reading code for a thrift struct.
"""
function readlongfield(p::CompactProtocol, k::Integer, t::Integer, id::Integer, T) 
    _readfield(readlongfieldheader, p, k, t, id, T)
end

"""
    readshort

A function to read the short form of a thrift struct.  Methods of this function are generated
by the `thriftstruct` macro.
"""
function readshort end

"""
    readlong

A function to read the long form of a thrift struct.  Methods of this function are generated
by the `thriftstruct` macro.
"""
function readlong end
